#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <math.h>

/* Definition of HildonStatusBarPluginFn_st */
#include <libhildondesktop/libhildondesktop.h>
#include <libosso.h>
#include <gtk/gtk.h>
// #include <cairo/cairo.h>
#include <gconf/gconf-client.h>
#include <gdk-pixbuf/gdk-pixdata.h>

#include "statusbarclock.h"

HD_DEFINE_PLUGIN(StatusbarClock,
                 statusbarclock,
                 STATUSBAR_TYPE_ITEM);

#define STATUSBARCLOCK_GET_PRIVATE(x) \
    (G_TYPE_INSTANCE_GET_PRIVATE ((x), TYPE_STATUSBARCLOCK, StatusbarClockPrivate))

#include "statusbarclock-face.h"
#include "statusbarclock-digiface.h"
#include "statusbarclock-mirrordigits.h"

#define COMPOSITE(a,b,v) ((guchar) (((guint16) (a) * (v) +\
                                     (guint16) (b) * (255 - (v)))\
                                    >> 8))

#define CLOCK_CENTER_X 20
#define CLOCK_CENTER_Y 19
#define CLOCK_RADIUS_M 11
#define CLOCK_RADIUS_H 7

#define CLOCK_DIGI_TOP    16
#define CLOCK_DIGI_HOUR_X  2
#define CLOCK_DIGI_MIN_X  22
#define CLOCK_DIGI_WIDTH   8
#define CLOCK_DIGI_HEIGHT 24

#define GCONF_CLOCK_MODE_DIR "/apps/osso/clock"
#define GCONF_CLOCK_MODE GCONF_CLOCK_MODE_DIR "/display-mode"

static void   statusbarclock_dispose (GObject     *object);

typedef struct _StatusbarClockPrivate StatusbarClockPrivate;
struct _StatusbarClockPrivate
{
  osso_context_t  *context;
  GtkWidget       *image;
  GdkPixdata      *face_pixdata;
  GdkPixdata      *digiface_pixdata;
  GdkPixdata      *digits_pixdata;
  guint            timeout_id;

  GConfClient     *gconf_client;
  guint            gconf_notify;

  GdkPixbuf       *digits;
  gboolean         analog;
  gboolean         dark;
};


static void
gdk_pixbuf_composite_pixel (GdkPixbuf *pixbuf,
                            int x, int y,
                            guchar red, guchar green, guchar blue,
                            guchar alpha)
{
  int width, height, rowstride, n_channels;
  guchar *pixels, *p;

  n_channels = gdk_pixbuf_get_n_channels (pixbuf);

  g_assert (gdk_pixbuf_get_colorspace (pixbuf) == GDK_COLORSPACE_RGB);
  g_assert (gdk_pixbuf_get_bits_per_sample (pixbuf) == 8);
  g_assert (gdk_pixbuf_get_has_alpha (pixbuf));
  g_assert (n_channels == 4);

  width = gdk_pixbuf_get_width (pixbuf);
  height = gdk_pixbuf_get_height (pixbuf);

  g_assert (x >= 0 && x < width);
  g_assert (y >= 0 && y < height);

  rowstride = gdk_pixbuf_get_rowstride (pixbuf);
  pixels = gdk_pixbuf_get_pixels (pixbuf);

  p = pixels + y * rowstride + x * n_channels;
  p[0] = COMPOSITE (red, p[0], alpha);
  p[1] = COMPOSITE (green, p[1], alpha);
  p[2] = COMPOSITE (blue, p[2], alpha);
}


static void
gdk_pixbuf_fill_color (GdkPixbuf *pixbuf,
                       guchar red, guchar green, guchar blue)
{
  int width, height, rowstride, n_channels;
  int x, y;
  guchar *pixels, *p;

  n_channels = gdk_pixbuf_get_n_channels (pixbuf);

  g_assert (gdk_pixbuf_get_colorspace (pixbuf) == GDK_COLORSPACE_RGB);
  g_assert (gdk_pixbuf_get_bits_per_sample (pixbuf) == 8);
  g_assert (gdk_pixbuf_get_has_alpha (pixbuf));
  g_assert (n_channels == 4);

  width = gdk_pixbuf_get_width (pixbuf);
  height = gdk_pixbuf_get_height (pixbuf);

  rowstride = gdk_pixbuf_get_rowstride (pixbuf);
  pixels = gdk_pixbuf_get_pixels (pixbuf);

  for (y = 0; y < height; y++)
    {
      for (x = 0; x < width; x++)
        {
          p = pixels + y * rowstride + x * n_channels;
          p[0] = red;
          p[1] = green;
          p[2] = blue;
        }
    }
}


static void
gdk_pixbuf_draw_line_aa (GdkPixbuf *pixbuf,
                         gint x0, gint y0, gint x1, gint y1,
                         guchar red, guchar green, guchar blue)
{
  gint dx, dy;
  gint x, y;
  gint slant;

  dx = x1 - x0;
  dy = y1 - y0;

  if (dx == 0 && dy == 0)
    {
      gdk_pixbuf_composite_pixel (pixbuf, x0, y0, red, green, blue, 255);
      return;
    }

  if (abs (dx) > abs (dy))
    {
      if (x1 < x0)
        {
          x = x0; x0 = x1; x1 = x;
          y = y0; y0 = y1; y1 = y;
          dx = x1 - x0;
          dy = y1 - y0;
        }

      slant = (dy << 8) / dx;

      y = y0 << 8;

      for (x = x0; x <= x1; x++)
        {
          gdk_pixbuf_composite_pixel (pixbuf,
                                      x, y >> 8,
                                      red, green, blue,
                                      255 - (y % 256));
          gdk_pixbuf_composite_pixel (pixbuf,
                                      x, (y >> 8) + 1,
                                      red, green, blue,
                                      y % 256);
          y += slant;
        }
    }
  else
    {
      if (y1 < y0)
        {
          x = x0; x0 = x1; x1 = x;
          y = y0; y0 = y1; y1 = y;
          dx = x1 - x0;
          dy = y1 - y0;
        }

      slant = (dx << 8) / dy;

      x = x0 << 8;

      for (y = y0; y <= y1; y++)
        {
          gdk_pixbuf_composite_pixel (pixbuf,
                                      x >> 8, y,
                                      red, green, blue,
                                      255 - (x % 256));
          gdk_pixbuf_composite_pixel (pixbuf,
                                      (x >> 8) + 1, y,
                                      red, green, blue,
                                      x % 256);
          x += slant;
        }
    }
}


static gboolean
statusbarclock_update_time (gpointer data)
{
  StatusbarClock *clock = STATUSBARCLOCK (data);
  StatusbarClockPrivate *pclock = STATUSBARCLOCK_GET_PRIVATE (clock);
  GdkPixbuf       *pixbuf = NULL;
  GError          *error  = NULL;
  gdouble          angle;
  struct timeval   tv;
  struct tm       *time_info;
  gchar           *tz_str;

  tz_str = g_strdup (g_getenv ("TZ"));
  g_unsetenv ("TZ");

  gettimeofday (&tv, NULL);
  time_info = localtime (&tv.tv_sec);

  if (tz_str)
    {
      g_setenv ("TZ", tz_str, FALSE);
      g_free (tz_str);
    }

  if (pclock->analog)
    {
      pixbuf = gdk_pixbuf_from_pixdata (pclock->face_pixdata,
                                        TRUE, &error);
      if (error) g_printerr ("%s\n", error->message);

      angle = (gdouble) time_info->tm_min / 30.0 * G_PI;
      gdk_pixbuf_draw_line_aa (pixbuf,
                               CLOCK_CENTER_X, CLOCK_CENTER_Y,
                               CLOCK_CENTER_X + sin (angle) * CLOCK_RADIUS_M + 0.5,
                               CLOCK_CENTER_Y - cos (angle) * CLOCK_RADIUS_M + 0.5,
                               80, 80, 80);

      angle = (gdouble) (time_info->tm_hour * 60 + time_info->tm_min) / 360.0 * G_PI;
      gdk_pixbuf_draw_line_aa (pixbuf,
                               CLOCK_CENTER_X, CLOCK_CENTER_Y,
                               CLOCK_CENTER_X + sin (angle) * CLOCK_RADIUS_H + 0.5,
                               CLOCK_CENTER_Y - cos (angle) * CLOCK_RADIUS_H + 0.5,
                               0, 0, 0);
    }
  else
    {
      pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE,
                               8, 40, 40);
      gdk_pixbuf_fill (pixbuf, 0x00000000);

      if (pclock->digits == NULL)
        {
          pclock->digits = gdk_pixbuf_from_pixdata (pclock->digits_pixdata,
                                                            TRUE, &error);
          if (error) g_printerr ("%s\n", error->message);
        }

      gdk_pixbuf_copy_area (pclock->digits,
                            (time_info->tm_hour / 10) * CLOCK_DIGI_WIDTH, 0,
                            CLOCK_DIGI_WIDTH, CLOCK_DIGI_HEIGHT,
                            pixbuf,
                            CLOCK_DIGI_HOUR_X, CLOCK_DIGI_TOP);

      gdk_pixbuf_copy_area (pclock->digits,
                            (time_info->tm_hour % 10) * CLOCK_DIGI_WIDTH, 0,
                            CLOCK_DIGI_WIDTH, CLOCK_DIGI_HEIGHT,
                            pixbuf,
                            CLOCK_DIGI_HOUR_X + CLOCK_DIGI_WIDTH, CLOCK_DIGI_TOP);

      gdk_pixbuf_copy_area (pclock->digits,
                            10 * CLOCK_DIGI_WIDTH, 0,
                            4, CLOCK_DIGI_HEIGHT,
                            pixbuf,
                            CLOCK_DIGI_HOUR_X + 2 * CLOCK_DIGI_WIDTH,
                            CLOCK_DIGI_TOP);

      gdk_pixbuf_copy_area (pclock->digits,
                            (time_info->tm_min / 10) * CLOCK_DIGI_WIDTH, 0,
                            CLOCK_DIGI_WIDTH, CLOCK_DIGI_HEIGHT,
                            pixbuf,
                            CLOCK_DIGI_MIN_X, CLOCK_DIGI_TOP);

      gdk_pixbuf_copy_area (pclock->digits,
                            (time_info->tm_min % 10) * CLOCK_DIGI_WIDTH, 0,
                            CLOCK_DIGI_WIDTH, CLOCK_DIGI_HEIGHT,
                            pixbuf,
                            CLOCK_DIGI_MIN_X + CLOCK_DIGI_WIDTH, CLOCK_DIGI_TOP);
      if (pclock->dark)
        gdk_pixbuf_fill_color (pixbuf, 0, 0, 0);
      else
        gdk_pixbuf_fill_color (pixbuf, 255, 255, 255);
    }

  gtk_image_set_from_pixbuf (GTK_IMAGE (pclock->image),
                             pixbuf);
  g_object_unref (pixbuf);

  if (pclock->timeout_id)
    g_source_remove (pclock->timeout_id);

  pclock->timeout_id = g_timeout_add ((60 - time_info->tm_sec) * 1000 + 10,
                                      statusbarclock_update_time,
                                      clock);

  return FALSE;
}


static void
statusbarclock_timezone_update (gpointer data)
{
  StatusbarClock *clock = STATUSBARCLOCK (data);

  statusbarclock_update_time (clock);
}


static void
statusbarclock_show_date (GtkButton *button,
                          gpointer   data)
{
  StatusbarClock *clock = STATUSBARCLOCK (data);
  StatusbarClockPrivate *pclock = STATUSBARCLOCK_GET_PRIVATE (clock);
  gchar           message[50];
  struct timeval  tv;
  struct tm      *time_info;
  gchar          *tz_str;

  tz_str = g_strdup (g_getenv ("TZ"));
  g_unsetenv ("TZ");

  gettimeofday (&tv, NULL);
  time_info = localtime (&tv.tv_sec);

  strftime (message, G_N_ELEMENTS (message), "%A\n%x", time_info);

  if (tz_str)
    {
      g_snprintf (message, G_N_ELEMENTS (message), "TZ=\"%s\"", tz_str);

      g_setenv ("TZ", tz_str, FALSE);
      g_free (tz_str);
    }

  osso_system_note_infoprint (pclock->context,
                              message, NULL);
}


#if 0
static gboolean
statusbarclock_expose_event (GtkWidget *widget,
                             GdkEventExpose *event,
                             gpointer   data)
{
  struct timeval  tv;
  struct tm      *time_info;
  gchar          *tz_str;
  gdouble         angle;
  cairo_t *cr;

  cr = gdk_cairo_create (event->window);

  gdk_cairo_rectangle (cr, &event->area);
  cairo_clip (cr);
  cairo_translate (cr, event->area.x, event->area.y);

  cairo_translate (cr, event->area.width / 2, event->area.height / 2);
  cairo_scale (cr, MIN (event->area.width / 2, event->area.height / 2),
               -MIN (event->area.width / 2, event->area.height / 2));
  cairo_scale (cr, 1.0/75, 1.0/75);
  cairo_set_source_rgb (cr, 1, 1, 1);
  cairo_arc (cr, 0, 0, 60, 0, 6.2832);
  cairo_set_line_width (cr, 9);
  cairo_stroke (cr);

  tz_str = g_strdup (g_getenv ("TZ"));
  g_unsetenv ("TZ");

  gettimeofday (&tv, NULL);
  time_info = localtime (&tv.tv_sec);

  if (tz_str)
    {
      g_setenv ("TZ", tz_str, FALSE);
      g_free (tz_str);
    }

  cairo_save (cr);
  angle = (gdouble) time_info->tm_min / 30.0 * G_PI;
  cairo_rotate (cr, -angle);
  cairo_move_to (cr, 0, -14);
  cairo_line_to (cr, 0, 45);
  cairo_set_line_width (cr, 6);
  cairo_stroke (cr);
  cairo_restore (cr);

  cairo_save (cr);
  angle = (gdouble) (time_info->tm_hour * 60 + time_info->tm_min) / 360.0 * G_PI;
  cairo_rotate (cr, -angle);
  cairo_move_to (cr, 0, -10);
  cairo_line_to (cr, 0, 33);
  cairo_set_line_width (cr, 10);
  cairo_stroke (cr);
  cairo_restore (cr);

  cairo_destroy (cr);

  return TRUE;
}
#endif


static void
statusbarclock_sleep_wakeup (osso_hw_state_t *state,
                             gpointer         data)
{
  StatusbarClock *clock = STATUSBARCLOCK (data);
  StatusbarClockPrivate *pclock = STATUSBARCLOCK_GET_PRIVATE (clock);

  if (pclock->timeout_id)
    g_source_remove (pclock->timeout_id);
  pclock->timeout_id = 0;

  if (state->system_inactivity_ind == FALSE)
    /* this also reregisters the timeout */
    statusbarclock_update_time (clock);
}


static void
statusbarclock_gconf_notify (GConfClient *client,
                             guint cnxn_id,
                             GConfEntry *entry,
                             gpointer data)
{
  StatusbarClock *clock = STATUSBARCLOCK (data);
  StatusbarClockPrivate *pclock = STATUSBARCLOCK_GET_PRIVATE (clock);
  GConfValue  *value;

  pclock->analog = TRUE;

  value = gconf_entry_get_value (entry);

  if (value && gconf_value_get_int (value) == 1)
    pclock->analog = FALSE;

  if (pclock->timeout_id)
    g_source_remove (pclock->timeout_id);
  pclock->timeout_id = 0;

  statusbarclock_update_time (clock);
}


static void
statusbarclock_init (StatusbarClock *clock)
{
  StatusbarClockPrivate *pclock = STATUSBARCLOCK_GET_PRIVATE (clock);
  GError        *error = NULL;
  GtkWidget     *button = NULL;
  osso_return_t  ret;

  pclock->gconf_client = gconf_client_get_default ();

  pclock->analog = TRUE;

  if (gconf_client_get_int (pclock->gconf_client,
                            GCONF_CLOCK_MODE, NULL) == 1)
    pclock->analog = FALSE;

  pclock->context = osso_initialize ("statusbarclock", "0.7",
                                             FALSE, NULL);
  if (!pclock->context) g_printerr ("osso_initialize() failed\n");

  pclock->face_pixdata = g_new (GdkPixdata, 1);
  gdk_pixdata_deserialize (pclock->face_pixdata,
                           G_N_ELEMENTS (clock_face),
                           clock_face, &error);
  if (error)
    {
      g_printerr ("%s (clock_face)\n", error->message);
      g_error_free (error);
      error = NULL;
    }

  pclock->digiface_pixdata = g_new (GdkPixdata, 1);
  gdk_pixdata_deserialize (pclock->digiface_pixdata,
                           G_N_ELEMENTS (statusbarclock_digiface),
                           statusbarclock_digiface, &error);
  if (error)
    {
      g_printerr ("%s (statusbarclock_digiface)\n", error->message);
      g_error_free (error);
      error = NULL;
    }

  pclock->digits_pixdata = g_new (GdkPixdata, 1);
  gdk_pixdata_deserialize (pclock->digits_pixdata,
                           G_N_ELEMENTS (statusbarclock_mirrordigits),
                           statusbarclock_mirrordigits, &error);
  if (error)
    {
      g_printerr ("%s (statusbarclock_mirrordigits)\n", error->message);
      g_error_free (error);
      error = NULL;
    }

  button = gtk_button_new ();
  g_signal_connect (G_OBJECT (button), "clicked",
                    G_CALLBACK (statusbarclock_show_date),
                    clock);
  pclock->image = gtk_image_new ();

  gtk_container_add (GTK_CONTAINER (button), pclock->image);
  gtk_container_add (GTK_CONTAINER (clock),  button);

#if 0
  g_signal_connect_after (G_OBJECT (button), "expose-event",
                          G_CALLBACK (statusbarclock_expose_event),
                          clock);

  gtk_widget_add_events (GTK_WIDGET (button), GDK_EXPOSURE_MASK);
#endif

  ret = osso_time_set_notification_cb (pclock->context,
                                       statusbarclock_timezone_update,
                                       clock);

  if (ret != OSSO_OK)
    g_printerr ("osso_time_set_notification_cb() failed\n");

  ret = osso_hw_set_event_cb (pclock->context,
                              NULL,
                              statusbarclock_sleep_wakeup,
                              clock);

  if (ret != OSSO_OK)
    g_printerr ("osso_hw_set_event_cb() failed\n");

  gconf_client_add_dir (pclock->gconf_client,
                        GCONF_CLOCK_MODE_DIR,
                        GCONF_CLIENT_PRELOAD_NONE,
                        &error);
  if (error)
    {
      g_printerr ("%s (gconf_client_add_dir)\n", error->message);
      g_error_free (error);
      error = NULL;
    }

  pclock->gconf_notify = gconf_client_notify_add (pclock->gconf_client,
                                                  GCONF_CLOCK_MODE,
                                                  statusbarclock_gconf_notify,
                                                  clock,
                                                  NULL, &error);
  if (error)
    {
      g_printerr ("%s (gconf_client_notify_add)\n", error->message);
      g_error_free (error);
      error = NULL;
    }

  /* start ticking */
  statusbarclock_update_time (clock);

  gtk_widget_show_all (GTK_WIDGET (clock));
}


static void
statusbarclock_dispose (GObject *clock)
{
  StatusbarClockPrivate *pclock = STATUSBARCLOCK_GET_PRIVATE (clock);

  if (pclock->digits)
    {
      g_object_unref (pclock->digits);
      pclock->digits = NULL;
    }

  if (pclock->gconf_client)
    {
      if (pclock->gconf_notify)
        gconf_client_notify_remove (pclock->gconf_client,
                                    pclock->gconf_notify);
      gconf_client_remove_dir (pclock->gconf_client,
                               GCONF_CLOCK_MODE_DIR, NULL);
      g_object_unref (pclock->gconf_client);
      pclock->gconf_client = NULL;
    }

  if (pclock->context)
    {
      osso_deinitialize (pclock->context);
      pclock->context = NULL;
    }

  if (pclock->timeout_id)
    {
      g_source_remove (pclock->timeout_id);
      pclock->timeout_id = 0;
    }
}



static void
statusbarclock_class_init (StatusbarClockClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS(klass);
  object_class->dispose = statusbarclock_dispose;
  g_type_class_add_private (klass, sizeof (StatusbarClockPrivate));
}

